import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.HashMap;
import java.util.Map;
import java.util.Stack;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/*
 * Lauren Wang & James White
 * Ling 575, Professor Tjalve, Spring 2012
 */
public class G2P
{
    Conversions rules;

    boolean skipStage4 = false;     // Used in testing to see what is given to Stage 4: the English-to-SAMPA conversion.

    public G2P() throws FileNotFoundException
    {
        this(new File("LW"));
    }

    public G2P(File lw) throws FileNotFoundException
    {
        rules = new Conversions(lw);
    }

    public static void main(String args[])
    {
        try {
            // Set up a G2P converter using the language model data in the "LW" subdirectory.
            G2P g2p = new G2P();

            if (args.length > 0) {
                // Join up all the arguments into one string so no fussy quoting is necessary.
                String input = join(args);

                // Convert the given sentence and print it.
                String output = g2p.convert(input);
                System.out.println("input '" + input + "' -> output '" + output + "'");
            } else {
                System.out.println("Enter text - empty line to exit.");
                BufferedReader br = new BufferedReader(new InputStreamReader(System.in));
                String line = null;

                while (((line = br.readLine()) != null) && (line.length() > 0)) {
                    String output = g2p.convert(line);
                    System.out.println("input '" + line + "' -> output '" + output + "'");
                }
            }

        } catch (IOException e) {
            e.printStackTrace();
            System.err.println("File not found.  Is the LW subdirectory in current directory and populated properly?");
        }

    }

    // Join up this array of strings with a space as an interword separator.
    static String join(String[] words)
    {
        StringBuilder sb = new StringBuilder();
        for (String word : words) {
            if (sb.length() > 0) sb.append(' ');
            sb.append(word);
        }

        return sb.toString();
    }

    public String convert(String sentence)
    {
        // Replace any character we won't interpret with a space.
        // Characters that are kept are letters, single quote (for the possessive), digits, dollar sign, and period.
        sentence = sentence.replaceAll("[^a-zA-Z1-9'.,$]", " ");

        sentence = replaceContextualAbbreviations(sentence);

        sentence = replaceAbbreviations(sentence);

        StringBuilder sb = new StringBuilder();

        String[] words = sentence.split("\\s+");

        for (String word : words) {
            if (sb.length() > 0) sb.append(' ');
            sb.append(toSAMPA(word));
        }

        return sb.toString();
    }

    public String toSAMPA(String word)
    {
        word = word.toLowerCase();

        // The possessive S is indicated by "'s" or "s'".
        boolean possessive = word.endsWith("'s") || word.endsWith("s'");

        if (possessive) {
            // Trim off the possessive marker before further processing - we'll add it back on later.
            word = word.substring(0, word.lastIndexOf('\''));
        }

        // All punctuation handling is finished.  Anything that isn't a letter is zapped now.
        word = word.replaceAll("[^a-z]", "");

        // Finally do the actual English phonemic conversion to SAMPA, if there's anything left of our word.
        if (word.length() > 0){
            // Flag to skip the conversion from graphemes to phonemes for testing.
            if (!skipStage4) {
                // Do the main English grapheme to SAMPA phoneme conversion.
                String sampa = processConsonants(processForSilentE(word));

                if ((sampa == null) || (sampa.length() < 1)) {
                    // This string doesn't fit our phonological model.
                    // We'll just spell out it's letters.

                    word = replaceAbbreviations(join(word.split("")));

                    sampa = processConsonants(processForSilentE(word));
                }

                word = sampa;
            }

            if (possessive && (word.length() > 0)) {
                // Add on the possessive S.
                switch (word.charAt(word.length() - 1)) {
                    case 't' :
                        word = word + "s";
                        break;
                    case 'z' :
                    case 's' :
                        word = word + "Iz";
                        break;
                    default :
                        word = word + "z";
                }
            }
        }

        return word;
    }

    static Pattern DATE = Pattern.compile("(?i)(January|February|March|April|May|June|July|August|September|October|November|December)" +
            "\\s+(\\d++)\\S*(?:\\s+(\\d{4}))?");

    String replaceContextualAbbreviations(String sentence)
    {
        // Context-sensitive substitutions go here.
        // These are things like "Dr. CapitalizedWord" -> "Doctor CapitalizedWord" and "CapitalizedWord Dr." -> "CapitalizedWord Drive".

        // The (regex, replacement) pairs are in the LW/abbreviations.txt file.
        // If the regex uses groups then only the first group will be replaced, otherwise the entire match is replaced.

        for (Map.Entry<String, String> abbreviation : rules.abbreviations.entrySet()) {
            String rep = abbreviation.getValue();
            String pat = abbreviation.getKey();

            // Contextual abbreviation replacement patterns are those that have a source group reference ($<group#>).
            if (rep.contains("$")) {
                Pattern pattern = Pattern.compile(pat);
                Matcher matcher = pattern.matcher(sentence);

                sentence = matcher.replaceAll(rep);
            }
        }

        // Ordinal day numbers for months are also contextual, but the replacement isn't a simple string.
        {
            Matcher matcher = DATE.matcher(sentence);

            StringBuffer sb = null;

            while (matcher.find()) {
                if (sb == null) sb = new StringBuffer();

                String replacement = matcher.group(1) + " " + numberToOrdinal(Integer.valueOf(matcher.group(2)));

                String year = matcher.group(3);
                if (year != null) {
                    replacement = replacement + " " + year.substring(0, 2) + " " + year.substring(2);
                }

                matcher.appendReplacement(sb, replacement);
            }

            if (sb != null) {
                matcher.appendTail(sb);
                sentence = sb.toString();
            }
        }

        return sentence;
    }

    // Mappings from cardinal number string suffix to ordinal number suffix.
    static Map<String, String> cardinalToOrdinal = new HashMap<String, String>() {
        {
            put("one", "first");
            put("two", "second");
            put("three", "third");
            put("four", "fourth");
            put("five", "fifth");
            put("six", "sixth");
            put("n", "nth");            // seven, ten, eleven, *teen
            put("eight", "eighth");
            put("nine", "ninth");
            put("twelve", "twelfth");
            put("y", "ieth");           // twenty, thirty, ...
        }
    };

    String numberToOrdinal(int number)
    {
        String cardinal = numberToCardinal(number);

        for (String key : cardinalToOrdinal.keySet()) {
            if (cardinal.endsWith(key)) {
                cardinal = cardinal.substring(0, cardinal.length() - key.length()) + cardinalToOrdinal.get(key);
                break;
            }
        }

        return cardinal;
    }

    static String[] tens = { "", "", "twenty", "thirty", "forty", "fifty", "sixty", "seventy", "eighty", "ninety" };

    static String[] ones = {
        "zero", "one", "two", "three", "four", "five", "six", "seven", "eight", "nine"
        , "ten", "eleven", "twelve", "thirteen", "fourteen", "fifteen", "sixteen", "seventeen", "eighteen", "nineteen"
    };

    private String numberToCardinal(int number)
    {
        StringBuilder cardinal = new StringBuilder();

        if (number >= 1000000) {
            cardinal.append(' ');
            cardinal.append(numberToCardinal(number / 1000000));
            cardinal.append(" million");
            number %= 1000000;
        }

        if (number >= 1000) {
            cardinal.append(' ');
            cardinal.append(numberToCardinal(number / 1000));
            cardinal.append(" thousand");
            number %= 1000;
        }

        if (number >= 100) {
            cardinal.append(' ');
            cardinal.append(numberToCardinal(number / 100));
            cardinal.append(" hundred");
            number %= 100;
        }

        if (number >= 20) {
            cardinal.append(' ');
            cardinal.append(tens[number / 10]);
            number %= 10;
        }

        if ((number > 0) || (cardinal.length() > 0)) {
            cardinal.append(' ');
            cardinal.append(ones[number]);
        }

        return cardinal.toString().trim();
    }

    static Pattern NUMBER = Pattern.compile("\\d+");

    // An abbreviation is ALL CAPS (with or w/o periods), mixed case letters with embedded periods, or no vowels (with or w/o periods).
    static Pattern ABBREVIATION = Pattern.compile("(?:\\p{Lu}\\.?)+|(?:\\p{L}*\\.){2,}|(?:[^aeiouyAEIOUY]\\.?)+");

    // A pronounceable abbreviation consists of one or more consonant-vowel pairs.
    static Pattern PRONOUNCEABLE_ABBREVIATION = Pattern.compile("(?:[\\p{L}&&[^aeiouy]][aeiouy])+[bdgklmnprstwx]?");

    String replaceAbbreviations(String sentence)
    {
        String[] words = sentence.split("\\s+");

        for (int i = 0; i < words.length; ++i) {
            for (Map.Entry<String, String> abbreviation : rules.abbreviations.entrySet()) {
                String rep = abbreviation.getValue();
                String pat = abbreviation.getKey();

                // A simple word abbreviation replacement will have no group back reference in the replacement pattern.
                if (!rep.contains("$")) {
                    if (words[i].equalsIgnoreCase(pat)) {
                        words[i] = rep;
                    }
                }
            }

            if (NUMBER.matcher(words[i]).matches()) {
                words[i] = numberToCardinal(Integer.valueOf(words[i]));
            }

            if (ABBREVIATION.matcher(words[i]).matches()) {
                // Just keep the letters now.
                words[i] = words[i].replaceAll("[^\\p{L}]", "").toLowerCase();

                // Should we pass it on to be pronounced as a word?
                if (!PRONOUNCEABLE_ABBREVIATION.matcher(words[i]).matches()) {
                    // Nope.  Split it up into letters and run the replacement again to get pronunciations for 'em.
                    words[i] = replaceAbbreviations(join(words[i].split("")));
                }
            }
        }

        return join(words);
    }

    public static final Pattern VOWEL = Pattern.compile("[aeiouy]+(gh|w|we)?");
    public static final Pattern PURE_VOWEL = Pattern.compile("[aeiou]+");
    public static final Pattern SILENT_E_GROUP = Pattern.compile("[^aeiou]*[aeiou]+[^aeiou]e(s|r|d)?$");
    public static final Pattern CONSONANT = Pattern.compile("[b-z&&[^eiou]]#?");
    public static final String J = "j";
    public static final String J_DIPTHONG_TAG = "*";
    public static final String SCHWA = "@";
    public static final int FUTURE = 2;
    public static final String END = "#";

    String processForSilentE(String word)
    {
        String result = "NOT PROCESSED";
        String backSyllable = "NO SILENT E TYPE VOWEL CLUSTER FOUND";
        String frontSyllable = "NO SILENT E TYPE VOWEL CLUSTER FOUND";
        String vowelCluster = "NO SILENT E TYPE VOWEL CLUSTER FOUND";
        Matcher m = SILENT_E_GROUP.matcher(word);
        if (m.find()) {//if ends in vowel+consonant + e/es/er/ed, then preceding vowel is tense.
            //convert the tense vowel to sampa
            backSyllable = m.group();
            frontSyllable = word.substring(0, m.start());
            Matcher mBack = PURE_VOWEL.matcher(backSyllable);
            mBack.find();
            vowelCluster = mBack.group();
            if (rules.silentE.containsKey(vowelCluster)) {
                backSyllable = mBack.replaceFirst(rules.silentE.get(vowelCluster));
            } else if (rules.general.containsKey(vowelCluster)) {
                backSyllable = mBack.replaceFirst(rules.general.get(vowelCluster));
            }
            //remove silent-e type ending and process any prefix bits as general.
            if (backSyllable.substring(backSyllable.length() - 3).equals("ces")
                    || backSyllable.substring(backSyllable.length() - 3).equals("ses")) {
                backSyllable = backSyllable.substring(0, backSyllable.length() - 3) + "s@z";
            } else if (backSyllable.substring(backSyllable.length() - 3).equals("zes")) {
                backSyllable = backSyllable.substring(0, backSyllable.length() - 3) + "z@z";
            } else if (backSyllable.substring(backSyllable.length() - 3).equals("cer")) {
                backSyllable = backSyllable.substring(0, backSyllable.length() - 3) + "s@r";
            } else if (backSyllable.substring(backSyllable.length() - 3).equals("ced")) {
                backSyllable = backSyllable.substring(0, backSyllable.length() - 3) + "st";
            } else if (backSyllable.substring(backSyllable.length() - 3).equals("ted")
                    || backSyllable.substring(backSyllable.length() - 3).equals("ded")) {
                backSyllable = backSyllable.substring(0, backSyllable.length() - 3) + "4@d";
            } else if (backSyllable.substring(backSyllable.length() - 2).equals("er")) {
                backSyllable = backSyllable.substring(0, backSyllable.length() - 2) + "@r";
            } else if (backSyllable.substring(backSyllable.length() - 2).equals("es")) {
                backSyllable = backSyllable.substring(0, backSyllable.length() - 2) + "s";
            } else if (backSyllable.substring(backSyllable.length() - 2).equals("ed")) {
                backSyllable = backSyllable.substring(0, backSyllable.length() - 2) + "d";
            } else if (backSyllable.substring(backSyllable.length() - 2).equals("ce")) {
                backSyllable = backSyllable.substring(0, backSyllable.length() - 2) + "ce";
            } else if (backSyllable.substring(backSyllable.length() - 2).equals("ge")) {
                backSyllable = backSyllable.substring(0, backSyllable.length() - 2) + "ge";
            } else if (backSyllable.substring(backSyllable.length() - 3).equals("the")) {
                backSyllable = backSyllable.substring(0, backSyllable.length() - 3) + "the";
            } else {
                backSyllable = backSyllable.substring(0, backSyllable.length() - 1);
            }//remove single e at end
            result = processGenerally(frontSyllable) + backSyllable;
        } else {
            result = processForTwoSyllable(word);
        }
        return result;
    }

    /*
      * bit of a misnomer--does one or two syllables
      */
    String processForTwoSyllable(String word)
    {
        String result = word;
        String vowelCluster = "";
        Matcher m = VOWEL.matcher(word);
        if (m.find()) {
            vowelCluster = m.group();
        }
        String[] consonants = VOWEL.split(word, 2);//breaks word after first vowel cluster.  split gives us the array of what the pattern is between.
        //if between nothing (not there), then array has length 1
        //if one match, then array has length 2 (final element may be "")
        //if n>1 match, then array has length max(n+1,second arg of split(.,.))
        //for(String s:consonants){System.out.println(s);}
        //NB first = consonants[0]+vowelCluster;
        String second = "";
        if (consonants.length == 2) {
            second = consonants[1];
        }
        String[] rest = VOWEL.split(second, 3);
        if (rest.length < 2) {//no more vowels beyond first syllable (so one syllable), so word is one-syllable
            if (second.equals("")) {//word is open syllable, so vowel in first is probably tense
                if (rules.twoSyllableTense.containsKey(vowelCluster)) {
                    result = m.replaceFirst(rules.twoSyllableTense.get(vowelCluster));
                } else if (rules.general.containsKey(vowelCluster)) {
                    result = m.replaceFirst(rules.general.get(vowelCluster));
                }
                //else result just stays same as word
            } else if (second.length() > 0 && second.substring(0, 1).equals("r")) {//rhotacized vowel
                if (rules.general.containsKey(vowelCluster + "r")) {
                    result = m.replaceFirst(rules.general.get(vowelCluster + "r"));
                }
            } else {//treat as lax
                if (rules.twoSyllableLax.containsKey(vowelCluster)) {
                    result = m.replaceFirst(rules.twoSyllableLax.get(vowelCluster));
                } else if (rules.general.containsKey(vowelCluster)) {
                    result = m.replaceFirst(rules.general.get(vowelCluster));
                }
                //else result just stays same as word
            }
        } else if (rest.length > 2) {//word is more than two syllables
            //treat entire word as generic case.
            //(would obviously be nicer if you had a stemmer and sent off the root to the two-syllable processor,
            //but there are a lot of things that would be nicer that I'm not doing either.)
            result = processGenerally(word);
        } else if (Pattern.matches("[^aeiouywhrl]", second.substring(1, 2)) || second.substring(0, 2).equals("ll") || second.substring(0, 2).equals("rr")) {//basically, two-syllable case with second letter of second not a vowel or w,y,h,r,l
            //first syllable's vowel probably lax and second syllable's a schwa.  redistribute consonants between first and second.
            if (second.substring(0, 1).equals("r") && rules.general.containsKey(vowelCluster + "r")) {
                vowelCluster = rules.general.get(vowelCluster + "r");
            } else if (rules.twoSyllableLax.containsKey(vowelCluster)) {
                vowelCluster = rules.twoSyllableLax.get(vowelCluster);
            } else if (rules.general.containsKey(vowelCluster)) {
                vowelCluster = rules.general.get(vowelCluster);
            }
            result = consonants[0] + vowelCluster + second.substring(0, 1);
            second = second.substring(1);
            if (Pattern.matches("[^aeiou](o|y)", second.substring(second.length() - 2))) {
                if (second.substring(second.length() - 1).equals("o")) {
                    second = second.substring(0, second.length() - 1) + "oU";
                } else {
                    second = second.substring(0, second.length() - 1) + "i";
                }
            } else {
                Matcher mSecond = PURE_VOWEL.matcher(second);
                if (mSecond.find()) {
                    second = mSecond.replaceFirst(SCHWA);
                }
            }
            result = result + second;
        } else {//two syllable case, with first syllable (probably) open and tense and second syllable schwa
            //syllables unchanged
            if (rules.twoSyllableTense.containsKey(vowelCluster)) {
                vowelCluster = rules.twoSyllableTense.get(vowelCluster);
            } else if (rules.general.containsKey(vowelCluster)) {
                vowelCluster = rules.general.get(vowelCluster);
            }
            result = consonants[0] + vowelCluster;
            if (Pattern.matches("[^aeiou](o|y)", second.substring(second.length() - 2))) {
                if (second.substring(second.length() - 1).equals("o")) {
                    second = second.substring(0, second.length() - 1) + "oU";
                } else {
                    second = second.substring(0, second.length() - 1) + "i";
                }
            } else {
                Matcher mSecond = PURE_VOWEL.matcher(second);
                if (mSecond.find()) {
                    second = mSecond.replaceFirst(SCHWA);
                }
            }
            result = result + second;
        }
        return result;
    }

    String processGenerally(String word)
    {
        if (word.equals("")) {
            return "";
        }

        String result = "NOT PROCESSED";
        String vowelCluster = "NO VOWEL CLUSTER FOUND";
        Matcher m = VOWEL.matcher(word);
        String[] consonants = VOWEL.split(word, -2);//don't get rid of "" at end

        result = consonants[0];
        int i = 1;
        while (m.find()) {
            vowelCluster = m.group();
            if (rules.general.containsKey(vowelCluster)) {
                result = result + rules.general.get(vowelCluster) + consonants[i];
            } else {
                result = result + vowelCluster + consonants[i];
            }
            i++;
        }
        return result;
    }

    String processConsonants(String word)
    {
        Stack<String> grapheme = new Stack<String>();
        grapheme.push(word.substring(word.length() - 1) + END);
        for (int i = word.length() - 2; i > -1; i--) {
            grapheme.push(word.substring(i, i + 1));
        }
        String result = "";
        while (!grapheme.empty()) {
            String current = grapheme.pop();
            if (current.equals(J) && !grapheme.empty()) {
                String checkMe = grapheme.pop();
                if (checkMe.equals(J_DIPTHONG_TAG)) {
                    result = result + current;
                    current = "";
                } else {
                    grapheme.push(checkMe);
                }
            }
            Matcher m = CONSONANT.matcher(current);
            if (m.find()) {
                String[] plus = new String[FUTURE + 1];
                plus[0] = m.group();
                for (int i = 1; i <= FUTURE; i++) {
                    plus[i] = "";
                }
                //while(!grapheme.isEmpty()){System.out.println(grapheme.pop());}
                Stack<String> saver = new Stack<String>();
                for (int i = 1; i <= Math.min(grapheme.size(), FUTURE); i++) {
                    saver.push(grapheme.pop());
                    plus[i] = plus[i - 1] + saver.peek();
                }
                while (!saver.empty()) {
                    grapheme.push(saver.pop());
                }//put them back
                boolean keepGoing = true;
                String replacement = null;
                replacement = plus[0];
                int index = -1;
                for (int i = FUTURE; i > -1 && keepGoing; i--) {
                    if (rules.consonants.containsKey(plus[i])) {
                        keepGoing = false;
                        index = i;
                        replacement = rules.consonants.get(plus[i]);
                    }
                }
                while (index > 0) {
                    index--;
                    grapheme.pop();
                }

                if (replacement == null) {
                    return "";
                }

                result = result + replacement;
            } else {//wave vowels through
                result = result + current;
            }
        }
        if (result.substring(result.length() - 1).equals(END)) {
            result = result.substring(0, result.length() - 1);
        }
        return result;
    }

}
